import pytest

from .singleton import singleton


@pytest.fixture(scope="function")
def TestSingletonClass():

    @singleton
    class NewSingletonClass:
        def __init__(self, value):
            self.value = value

    yield NewSingletonClass


def test_singleton_instance(TestSingletonClass):
    obj1 = TestSingletonClass(1)
    obj2 = TestSingletonClass(1)

    assert obj1 is obj2  # Both objects should be the same instance


def test_singleton_init_only_called_once(TestSingletonClass):
    obj1 = TestSingletonClass(10)
    obj2 = TestSingletonClass(20)

    assert obj1 is obj2  # Both objects should be the same instance

    # value should not have been updated from first instantiation
    assert obj1.value == 10
    assert obj2.value == 10


def test_singleton_second_call_ignores_args(TestSingletonClass):
    obj1 = TestSingletonClass(10)
    obj2 = TestSingletonClass()
    obj3 = TestSingletonClass(1, 2, 3)

    assert obj1 is obj2
    assert obj1 is obj3



"""SingleBooleanComponent class module."""

from .core import BaseComponent
from .core import SingleValueComponent

from .single_boolean_components import SingleBooleanComponent, Toggle


def test_is_single_value_subclass():
    assert issubclass(SingleBooleanComponent, SingleValueComponent)


def test_default_value():
    c = SingleBooleanComponent("id")

    assert not c.value


def test_subclasses():
    assert issubclass(Toggle, SingleBooleanComponent)


def test_gets_registered():
    cid = "id"
    c = Toggle(cid)
    assert BaseComponent.registry.get(cid) == c



"""Test main file"""

import time

import zmc

# print("zmc", dir(zmc))
# print()
# print("zmc.components", dir(zmc.components))
# print()
# print("zmc.components.core", dir(zmc.components.core))

graph = zmc.components.Graph("sdf")

with zmc.connect(debug=True, verbose=False):
    for i in range(1, 50 + 1):
        graph.plot([0] * i, [0] * i)
        print(i)
        time.sleep(0.01)


# print("websockets", dir(websockets))
# websockets.framing
# print("websockets", dir(websockets))


# try:
#     zmc.connect
# except:
#     print("no connect")

# try:
#     zmc.components
# except Exception:
#     print("no components")


# a = zmc.components.SingleBooleanComponent("asdf")
# b = zmc.components.button_component.ButtonComponent("d")
# c = zmc.components.BADTHING("d")


# base -> receiver / registry -> websocket -> connector_context -> sender



"""WebsocketServer class module."""

import asyncio
import json
import websockets

from zmc.components.core import BaseComponent
from zmc.components.core import ValueReceiverBaseComponent
from zmc.version import __version__

from .async_utils import cancellable, async_repeat
from .connector import Connector


__all__ = ["WebSocketServer"]


# User friendly error message regarding the fact that a component that does not
# allow for setting a value was sent a value from the Mission Control app.
SETTING_VALUE_ON_NON_SETTER_ERROR_TEMPLATE = (
    "Component id, '{}', was attached to a component that does not expect to "
    "receive any values. However, the Mission Control app is sending values to "
    "the component with this id.\nPlease verify that you are using the right "
    "id when initializing this component."
)

# Rate (in Hz) at which messages will be sent over websockets.
_THROTTLE_RATE = 61


class WebSocketServer(Connector):
    """Server that allows caller to open up a websocket on a separate thread.

    Args:
        - host: host as per similarly named asyncio.serve() function parameter
        - port: port as per similarly named asyncio.serve() function parameter
        - gui_id: id of the gui this server is going to connect to.
    """

    def __init__(self, host, port, gui_id):
        super().__init__()
        self.host = host
        self.port = port
        self.gui_id = gui_id

        # Server/websocket related attributes
        self.__server = None
        self.__websockets = set()

        # Queued data related attirbutes
        self.__data_component_queue = asyncio.Queue()
        self.__send_queued_data_task = None

    def _reset(self):
        self.__server = None
        self.__websockets.clear()
        self.__data_component_queue = asyncio.Queue()
        self.__send_queued_data_task = None

    @cancellable
    async def _send_all_component_data(self, components):
        data_event = {
            "type": "send-all-data",
            "guiId": self.gui_id,
            "components": [
                {
                    "componentId": component.id,
                    "data": component.data,
                }
                for component in components.values()
            ],
            "pyVersion": __version__,
        }
        json_data_event = json.dumps(data_event)
        websockets_set = self.__websockets.copy()
        for websocket in websockets_set:
            try:
                await websocket.send(json_data_event)
            except websockets.ConnectionClosed:
                pass

    @cancellable
    @async_repeat(1 / _THROTTLE_RATE)
    async def _send_queued_component_data(self):
        items = {}
        while not self.__data_component_queue.empty():
            component = await self.__data_component_queue.get()
            items[component.id] = component
        if items:
            await self._send_all_component_data(items)

    async def _start_send_queued_data_task(self):
        if (
            self.__send_queued_data_task is None
            or self.__send_queued_data_task.done()
        ):
            self.__send_queued_data_task = asyncio.create_task(
                self._send_queued_component_data()
            )

    async def _send_component_data(self, component):
        """Queues up component data to be sent."""
        # Add component data to the queue
        await self.__data_component_queue.put(component)
        # Start the queued data sending loop if it hasn't been already.
        await self._start_send_queued_data_task()

    # TODO: should refactor this to be a little more palettable
    async def _websocket_consumer(self, websocket):
        async for message in websocket:
            event = json.loads(message)
            # TODO: handle app versions (for future proofing as things change)
            _app_version = event["app_version"]
            if event["type"] == "set-value":
                component_id = event["component_id"]
                component = BaseComponent.registry.get(component_id, None)
                if component is None:
                    continue  # Didn't find component. Do nothing and skip.
                if not isinstance(component, ValueReceiverBaseComponent):
                    raise ValueError(
                        SETTING_VALUE_ON_NON_SETTER_ERROR_TEMPLATE.format(
                            component.id
                        )
                    )
                component.receive_value(event.get("value", None))
            if event["type"] == "initialize":
                components = event["components"]
                for component_json in components:
                    component_id = component_json["component_id"]
                    component = BaseComponent.registry.get(component_id, None)
                    if component is None:
                        continue  # Didn't find component. Do nothing and skip.
                    value = component_json.get("value", None)
                    if not value:
                        continue  # No value to update.
                    if not isinstance(component, ValueReceiverBaseComponent):
                        raise ValueError(
                            SETTING_VALUE_ON_NON_SETTER_ERROR_TEMPLATE.format(
                                component.id
                            )
                        )
                    component.receive_value(value)
                self.is_connected = True  # Signal that a connection is ready
            if event["type"] == "kill-func":
                self._exit_original_thread()

    async def _websocket_handler(self, websocket, _):
        self.__websockets.add(websocket)
        consumer_task = asyncio.create_task(self._websocket_consumer(websocket))

        done, pending = await asyncio.wait(
            [consumer_task], return_when=asyncio.FIRST_COMPLETED
        )

        for task in pending:
            task.cancel()

        self.__websockets.discard(websocket)
        if not self.__websockets and self.is_connected:
            # If there are no websockets left, then there is no connection.
            self.is_connected = False

        for task in done:
            try:
                error = task.exception()
                if error is not None:
                    self._handle_thread_exception(error)
            except asyncio.CancelledError:
                pass

    async def _start_connector(self):
        self.__server = await websockets.serve(
            self._websocket_handler, self.host, self.port
        )

    async def _stop_connector(self):
        if self.__server is not None and self.__server.is_serving():
            self.__server.close()
            await self.__server.wait_closed()

        self._reset()



from .base_component import BaseComponent
from .single_value_component import SingleValueComponent
from .value_receiver_base_component import ValueReceiverBaseComponent


def test_is_base_subclass():
    assert issubclass(SingleValueComponent, BaseComponent)
    assert issubclass(SingleValueComponent, ValueReceiverBaseComponent)


def test_default_value():
    val = 4  # Arbitrary value.
    c1 = SingleValueComponent("id", val)

    assert c1.value == val


def test_set_value():
    c1 = SingleValueComponent("id", 5)
    val = 9  # Arbitrary value.
    c1.receive_value({"value": val})

    assert c1.value == val



"""singleton decorator utils module."""

import functools


def singleton(cls):
    """Make a class a Singleton class.

    The __new__ and __init__ functions will only be called the first time the
    class is instantiated. All other times, the first instance will be returned
    without making any calls to the class itself.
    """

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if wrapper_singleton.instance is None:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton



"""ZMC package module."""

# from .connect.connector_context import ConnectorContext as connect

# from .components import *

from .version import __version__

# __all__ = ["__version__", "connect"]
del version



"""SingleNumberComponent class module."""

from .core import SingleValueComponent


__all__ = ["Slider", "NumericInput"]


class SingleNumberComponent(SingleValueComponent):
    """Number receiver class which contains a single value."""

    def __init__(self, component_id):
        super().__init__(component_id, 1)


class Slider(SingleNumberComponent):
    """Slider class which contains a single value.

    A slider has a value of a single number that can be accessed as an
    attribute: `slider.value`.
    """


class NumericInput(SingleNumberComponent):
    """Numeric class, representing a freeform or dropdown number component.

    The class has a single value that can be accessed as an attribute:
    `numeric_input.value`.
    """



"""Test main file"""

import zmc

print(dir(zmc))

# try:
#     zmc.connect
# except:
#     print("no connect")

try:
    zmc.components
except Exception:
    print("no components")


# a = zmc.components.SingleBooleanComponent("asdf")
# b = zmc.components.button_component.ButtonComponent("d")
# c = zmc.components.BADTHING("d")



import pytest

from .core import _BaseComponentMeta


@pytest.fixture(autouse=True, scope="function")
def clear_base_component_class_registry():
    # The BaseComponent registry should already be empty but, just in case, we
    # clear both before and after the test is run.
    _BaseComponentMeta._BaseComponentMeta__REGISTRY.clear()
    yield
    _BaseComponentMeta._BaseComponentMeta__REGISTRY.clear()



import pytest

from .base_component import BaseComponent
from .component_instance_registry import ComponentInstanceRegistry


@ComponentInstanceRegistry
class Component(BaseComponent):
    def __init__(self, component_id, value=None):
        super().__init__(component_id)
        self.value = value

    def get_value(self):
        return self.value


@pytest.fixture(autouse=True, scope="function")
def clear_registry():
    ComponentInstanceRegistry.clear_registry()


def test_instance_type_is_preserved():
    c1 = Component("id1")

    assert isinstance(c1, Component)
    assert isinstance(c1, BaseComponent)


def test_class_type_is_preserved():
    assert issubclass(Component, BaseComponent)
    assert not issubclass(Component, ComponentInstanceRegistry)


def test_instance_gets_registered():
    id1 = "id1"
    id2 = "id2"
    c1 = Component(id1)
    c2 = Component(id2)

    assert ComponentInstanceRegistry.get_instance(id1) == c1
    assert ComponentInstanceRegistry.get_instance(id2) == c2
    assert ComponentInstanceRegistry.get_instance("not registered id") is None


def test_non_unique_id_raises():
    cid = "id1"
    Component(cid)
    with pytest.raises(ValueError, match="already.*instantiated with id.*" + cid):
        Component(cid)


def test_decorate_non_component_subclass_raises():
    with pytest.raises(
        ValueError, match="can only decorate subclasses of BaseComponent"
    ):

        @ComponentInstanceRegistry
        class BadComponent:
            pass


def test_instance_initialization_preserved():
    value = 6  # arbitrary value
    c1 = Component("id", value)

    assert c1.get_value() == value



"""ComponentInstanceRegistry module."""

import functools

from .base_component import BaseComponent

__all__ = ["ComponentInstanceRegistry"]


# TODO: handle unregistering instances?
# TODO: decorator around BaseComponent to automatically add subclasses to registry?
class ComponentInstanceRegistry:
    """Registry of all `BaseComponent` instances created in the program.

    Any instance of a component class (i.e. subclass of `BaseComponent`)
    decorated  with `@ComponentInstanceRegistry` will automatically register
    instances in a shared registry. These instances will be able to be accessed
    from the registry from anywhere in the program that has access to it.

    When an instance is created, it is added automatically to the registry with
    no additional work needed from the user. If another instance with the same
    id has already been registered, an error will be raised.

    To enroll a subclass of `BaseComponent` to the registry, simply decorate it
    with `@ComponentInstanceRegistry`:
    ```
    @ComponentInstanceregistry
    class SomeComponent(BaseComponent):
        ...
    ```

    Decorating a class that is not a subclass of `BaseComponent` will result in
    a `ValueError` being raised.
    """

    __instances = {}

    def __new__(cls, component_cls):
        if not issubclass(component_cls, BaseComponent):
            raise ValueError(
                "ComponentInstanceRegistry can only decorate subclasses of "
                f"BaseComponent, given: {component_cls}"
            )
        original_init = component_cls.__init__

        @functools.wraps(original_init)
        def wrapped_init(instance, *args, **kwargs):
            original_init(instance, *args, **kwargs)
            ComponentInstanceRegistry.__register_instance(instance)

        component_cls.__init__ = wrapped_init
        return component_cls

    @staticmethod
    def __register_instance(instance):
        if instance.id in ComponentInstanceRegistry.__instances:
            raise ValueError(
                "There is already a component instantiated with id, given:", instance.id
            )
        ComponentInstanceRegistry.__instances[instance.id] = instance

    @staticmethod
    def get_instance(component_id):
        """Get a component instance based on its id.

        Args:
            - component_id: id of the component being retrieved.

        Returns:
            The component instance if there is one with component_id as its id,
                otherwise None.
        """
        return ComponentInstanceRegistry.__instances.get(component_id, None)

    @staticmethod
    def clear_registry():
        """Clear the registry. This removes all registered instances."""
        ComponentInstanceRegistry.__instances.clear()



"""Graph class module."""

from .core import DataSenderBaseComponent


__all__ = ["Graph"]


class Graph(DataSenderBaseComponent):
    """Graph component class."""

    def __init__(self, component_id):
        super().__init__(component_id)
        self._x = []
        self._y = []

    @property
    def data(self):
        return {
            "lines": [
                {
                    "x": list(self._x),
                    "y": list(self._y),
                }
            ]
        }

    def append_data(self, x, y):
        """Add a x, y pair to the graph and send data via server."""
        self._x.append(x)
        self._y.append(y)
        self._send_data()

    def plot(self, x, y):
        """Replace graph data entirely and send it."""
        self._x = x
        self._y = y
        self._send_data()



import pytest

from .instance_registry import InstanceRegistry


class ClassWithId:
    def __init__(self, instance_id):
        self.id = instance_id


class ClassWithName:
    def __init__(self, name):
        self.name = name


def test_registering_instances():
    registry = InstanceRegistry()
    id1 = "id1"
    id2 = "id2"
    c1 = ClassWithId(id1)
    c2 = ClassWithId(id2)
    registry.register(c1)
    registry.register(c2)

    assert registry.get_instance(id1) == c1
    assert registry.get_instance(id2) == c2
    assert registry.get_instance("not registered id") is None
    assert len(registry) == 2


def test_custom_id_fn():
    registry = InstanceRegistry(id_fn=lambda x: x.name)
    name = "name"
    id1 = "id"
    c1 = ClassWithName(name)
    registry.register(c1)

    assert registry.get_instance(id1) is None
    assert registry.get_instance(name) == c1


def test_deregistering_instance():
    registry = InstanceRegistry()
    id1 = "id1"
    c1 = ClassWithId(id1)
    registry.register(c1)
    assert registry.get_instance(id1) == c1

    registry.deregister(id1)
    assert registry.get_instance(id1) is None


def test_clear_registry():
    registry = InstanceRegistry()
    id1 = "id1"
    id2 = "id2"
    c1 = ClassWithId(id1)
    c2 = ClassWithId(id2)
    registry.register(c1)
    registry.register(c2)
    assert registry.get_instance(id1) == c1
    assert registry.get_instance(id2) == c2

    registry.clear()
    assert len(registry) == 0


def test_get_dict():
    registry = InstanceRegistry()
    id1 = "id1"
    id2 = "id2"
    c1 = ClassWithId(id1)
    c2 = ClassWithId(id2)
    registry.register(c1)
    registry.register(c2)
    assert registry.get_dict() == {id1: c1, id2: c2}


def test_asserting_type():
    registry = InstanceRegistry(instance_type=ClassWithId)
    id1 = "id1"
    name2 = "name2"
    c1 = ClassWithId(id1)
    c2 = ClassWithName(name2)
    registry.register(c1)
    with pytest.raises(
        ValueError,
        match=f"Expected.*instance of.*{ClassWithId}.*given.*{ClassWithName}",
    ):
        registry.register(c2)


def test_non_unique_id_raises():
    registry = InstanceRegistry()
    cid = "id1"
    c1 = ClassWithId(cid)
    registry.register(c1)
    c2 = ClassWithId(cid)
    with pytest.raises(
        ValueError, match="already.*instance with the same id.*" + cid
    ):
        registry.register(c2)


def test_no_positional_args():
    with pytest.raises(TypeError):
        # pylint: disable=too-many-function-args
        InstanceRegistry(lambda x: x.name)



"""InstanceRegistry module."""

__all__ = ["InstanceRegistry"]


# TODO: consider emulating dict `get` functionality. Might just mean implement a
#       few special functions and removing `get_instance`. If we do this, we
#       should maybe remove `get_dict` func and replace it with `copy`.
class InstanceRegistry:
    """Registry for instances.

    Any instance registered is associated with an id that is retrieved by
    calling `id_fn`. This function is either passed in during registry
    initialization or defaults to calling `instance.id`. If a type is passed in
    at initialization, then all registered intances will be asserted to be
    instances of said type.

    All registered instances are required to have unique ids. Trying to register
    an instance with an id that has already been registered will result in an
    error being raised.

    Args:
        - id_fn: function that returns id of instance. Defaults to `instance.id`
        - instance_type: expected type of instances.
    """

    def __init__(self, *, id_fn=lambda x: x.id, instance_type=None):
        self.__instances = {}
        self.__id_fn = id_fn
        self.__instance_type = instance_type

    def register(self, instance):
        """Register an instance.

        Args:
            - instance: the instance being registered

        Raises:
            ValueError: an instance with that id has already been registered.
            ValueError: if  `instance_type` was passed in but type of instance
                does not match.
        """
        if self.__instance_type and not isinstance(
            instance, self.__instance_type
        ):
            raise ValueError(
                f"Expected to register an instance of {self.__instance_type}, "
                f"given: {instance.__class__}"
            )

        instance_id = self.__id_fn(instance)
        if instance_id in self.__instances:
            raise ValueError(
                "There is already an instance with the same id, given:",
                instance_id,
            )
        self.__instances[instance_id] = instance

    def deregister(self, instance_id):
        """Deregister an instance if it is registered.

        Args:
            - instance_id: the id of an instance already registered
        """
        if instance_id in self.__instances:
            del self.__instances[instance_id]

    def get_instance(self, instance_id, default=None):
        """Get an instance based on its id.

        Args:
            - instance_id: Id of the instance being retrieved.
            - default: What to return if there is no instance with id
                instance_id

        Returns:
            The instance if there is one with an id of instance_id, otherwise
                default (which is None unless explicitely set).
        """
        return self.__instances.get(instance_id, default)

    def clear(self):
        """Clear the registry. This removes all registered instances."""
        self.__instances.clear()

    def get_dict(self):
        """Dict representation of the registry."""
        return dict(self.__instances)

    def __len__(self):
        return len(self.__instances)



import pytest

from .base_component import BaseComponent


class Component(BaseComponent):
    pass


class Component2(BaseComponent):
    pass


class ComponentWithError(BaseComponent):
    def __init__(self, instance_id):
        super().__init__(instance_id)
        raise RuntimeError("Oh no!")


def test_id_available():
    cid = "id1"
    c1 = Component(cid)

    assert c1.id is cid


def test_invalid_id_raises():
    with pytest.raises(
        ValueError, match=f"Component id must be a non empty string.*{None}"
    ):
        Component(None)
    with pytest.raises(
        ValueError, match=f"Component id must be a non empty string.*{9}"
    ):
        Component(9)
    with pytest.raises(
        ValueError, match="Component id must be a non empty string.*"
    ):
        Component("")


def test_registry():
    id1 = "id1"
    id2 = "id2"
    c1 = Component(id1)
    c2 = Component(id2)

    assert BaseComponent.registry == {id1: c1, id2: c2}


def test_without_super_init_call():
    class ComponentWithoutSuperInit1(BaseComponent):
        """A very hacky class that does not call the super __init__ method.

        For testing purposes, the class has a valid `self.id` so that it can be
        registered.
        """

        # pylint: disable=super-init-not-called
        def __init__(self, instance_id):
            # pylint: disable=invalid-name
            self._BaseComponent__id = instance_id

    cid = "id"
    c = ComponentWithoutSuperInit1(cid)
    assert BaseComponent.registry.get(cid) == c

    class ComponentWithoutSuperInit2(BaseComponent):
        """A class that does not call the super __init__ method."""

        # pylint: disable=super-init-not-called
        def __init__(self):
            pass

    with pytest.raises(AttributeError):
        ComponentWithoutSuperInit2()


def test_error_in_init_does_not_register():
    cid = "id"
    with pytest.raises(RuntimeError):
        ComponentWithError(cid)

    assert cid not in BaseComponent.registry


def test_non_unique_id_raises():
    cid = "id1"
    Component(cid)
    # Instance from same class and same id, raises
    with pytest.raises(
        ValueError, match="already.*instance with the same id.*" + cid
    ):
        Component(cid)
    # Instance from different subclass and same id, raises
    with pytest.raises(
        ValueError, match="already.*instance with the same id.*" + cid
    ):
        Component2(cid)



"""SingleBooleanComponent class module."""

from .core import BaseComponent
from .core import SingleValueComponent

from .single_text_components import TextInput, SingleTextComponent


def test_is_single_value_subclass():
    assert issubclass(SingleTextComponent, SingleValueComponent)


def test_default_value():
    c = SingleTextComponent("id")

    assert c.value == ""


def test_subclasses():
    assert issubclass(TextInput, SingleTextComponent)


def test_gets_registered():
    cid = "id"
    c = TextInput(cid)
    assert BaseComponent.registry.get(cid) == c



"""."""

import zmc

print("zmc", dir(zmc))
print()

# print("zmc connect", zmc.connect.__name__)
# print()

print("zmc components", dir(zmc.components))
print()

print("zmc components core", dir(zmc.components.core))
print()

print("zmc version", zmc.version.__version__)
print()

print("zmc utils", dir(zmc.utils))
print()



"""SingleTextComponent class module."""

from .core import SingleValueComponent


__all__ = ["TextInput"]


class SingleTextComponent(SingleValueComponent):
    """Text receiver class which contains a single value."""

    def __init__(self, component_id):
        super().__init__(component_id, "")


class TextInput(SingleTextComponent):
    """Text class, representing a freeform or dropdown text component.

    The class has a single value that can be accessed as an attribute:
    `text_input.value`.
    """



import re
import pytest

from zmc.connect.connector_context import context_connector

from .base_component import BaseComponent
from .data_sender_base_component import (
    DataSenderBaseComponent,
    DATA_SENT_OUTSIDE_OF_CONTEXT_ERROR,
)


class Component(DataSenderBaseComponent):

    def __init__(self, component_id):
        super().__init__(component_id)
        self._data = "dummy data"

    @property
    def data(self):
        return self._data  # pragma: no cover

    def send_data(self):
        self._send_data()


@pytest.fixture
def mock_context_server(mocker):
    mock_instance = mocker.Mock()

    token = context_connector.set(mock_instance)
    yield mock_instance
    context_connector.reset(token)


def test_is_base_subclass():
    assert issubclass(DataSenderBaseComponent, BaseComponent)


def test_is_abstract():
    with pytest.raises(TypeError, match="abstract class"):
        # pylint:disable=abstract-class-instantiated
        DataSenderBaseComponent("id")


def test_outside_of_context_raises():
    c = Component("id")

    with pytest.raises(
        RuntimeError, match=re.escape(DATA_SENT_OUTSIDE_OF_CONTEXT_ERROR)
    ):
        c.send_data()


def test_inside_context_calls_send(mock_context_server):
    component_id = "id"
    c = Component(component_id)
    c.send_data()

    mock_context_server.send_component_data.assert_called_once()
    mock_context_server.send_component_data.assert_called_once_with(c)



"""Component implementations.

Any component implementation should be exposed through here.
"""

from .button_component import *
from .graph_component import *
from .single_boolean_components import *
from .single_text_components import *
from .single_number_components import *


# Remove files which are added to the namespace because of imports but are not
# meant to be accessed directly by the user.
# pylint:disable=undefined-variable
# mypy: ignore-errors
del (
    button_component,
    graph_component,
    single_boolean_components,
    single_text_components,
    single_number_components,
)



"""Util functions for async functionality."""

import asyncio
import functools
import inspect
import time


def cancellable(coro_fn):
    """Convenience decorator that makes a coroutine function cancellable."""

    if not inspect.iscoroutinefunction(coro_fn):
        raise ValueError("cancellable can only decorate async functions.")

    @functools.wraps(coro_fn)
    async def cancellable_wrapper(*args, **kwargs):
        try:
            return await coro_fn(*args, **kwargs)
        except asyncio.CancelledError:
            pass  # don't throw anything when canceled.

    return cancellable_wrapper


# TODO: assert no event loop is running? It is sync after all
# TODO: gather will hang if loop is running on run_until_complete.
def cancel_loop_tasks(loop):
    """Cancels all tasks in a given loop running on a separate thread."""

    async def cancel_all_tasks():
        # Gather all tasks that are not this current one and cancel them.
        tasks = [
            task
            for task in asyncio.all_tasks(loop)
            if task is not asyncio.current_task()
        ]
        for task in tasks:
            task.cancel()
        # Wait for all tasks to be canceled before returning.
        await asyncio.gather(*tasks, return_exceptions=True)

    # Only try and cancel tasks if the loop is running.
    if not loop.is_running():
        return

    future = asyncio.run_coroutine_threadsafe(cancel_all_tasks(), loop)
    future.result()


def async_repeat(interval):
    """Convenience decorator that repeats a coroutine at regular intervals.

    The precision of this function is within a few milliseconds of the target
    interval, especially averaged over multiple repetitions.

    Args:
        - interval: How long to wait in between repetitions (in s)
    """

    def repeat_decorator(coro_fn):
        """Convenience decorator that makes a coroutine function cancellable."""

        if not inspect.iscoroutinefunction(coro_fn):
            raise ValueError("async_repeat can only decorate async functions.")

        @functools.wraps(coro_fn)
        async def repeat_wrapper(*args, **kwargs):
            time_delta, start = 0, time.time()
            while True:
                remaining_time = max(interval - time_delta, 0)
                await asyncio.sleep(remaining_time)
                prev_start = start
                start = time.time()
                # Calculate how much longer the wait took compared to the target
                # interval. asyncio.sleep is not extremely precise so this helps
                # keep things more or less accurate over time.
                extra_sleep_delay = start - prev_start - interval
                await coro_fn(*args, **kwargs)
                end = time.time()
                # `time_delta` is however long it has been since the end of the
                # previous sleep call plus however much the previous sleep call
                # overshot the taret interval.
                time_delta = end - start + extra_sleep_delay

        return repeat_wrapper

    return repeat_decorator



import pytest

from .base_component import BaseComponent
from .value_receiver_base_component import ValueReceiverBaseComponent


# TODO: add tests for callbacks


def test_is_base_subclass():
    assert issubclass(ValueReceiverBaseComponent, BaseComponent)


def test_is_abstract():
    with pytest.raises(TypeError, match="abstract class"):
        ValueReceiverBaseComponent("id")


def test_impl():
    class Component(ValueReceiverBaseComponent):
        def _set_value(self, value):
            self.value = value

    c = Component("id")
    value = 9
    c.receive_value(value)
    assert c.value == value



"""BaseComponentValueSender class module."""

from abc import abstractmethod


from .base_component import BaseComponent

__all__ = ["DataSenderBaseComponent"]

# User friendly error message regarding the fact that the user attempted to send
# data outside a server context.
DATA_SENT_OUTSIDE_OF_CONTEXT_ERROR = (
    "Attempted to send data to Mission Control app outside of the "
    "`zmc.connect()` context which is not possible. Make sure to run your code "
    "inside of a `with zmc.connect():` context scope or run it in a function "
    "that has been decorated with `@zmc.connect()`."
)


class DataSenderBaseComponent(BaseComponent):
    """Base class for any component that will send data."""

    @property
    @abstractmethod
    def data(self):
        """Data corresponding to component.

        This function should be called in a separate thread so that whatever
        preprocessing is needed does not block calls or affect timing of the
        user's code.
        """

    def _send_data(self):
        # pylint:disable=import-outside-toplevel
        from zmc.connect.connector_context import context_connector

        server = context_connector.get()
        if server is None:
            raise RuntimeError(DATA_SENT_OUTSIDE_OF_CONTEXT_ERROR)
        server.send_component_data(self)



"""ButtonComponent class module."""

from ..component_instance_registry import ComponentInstanceRegistry
from ..value_receiver_base_component import ValueReceiverBaseComponent


__all__ = ["ButtonComponent"]


@ComponentInstanceRegistry
class ButtonComponent(ValueReceiverBaseComponent):
    """Button class that calls a function whenever it is clicked."""

    def _set_value(self, _):
        pass  # No value but this is still needed to make the class concrete.



import asyncio
import inspect
import threading
import pytest

from .async_utils import cancellable, cancel_loop_tasks


async def async_fn():
    await asyncio.sleep(5)


@cancellable
async def cancellable_async_fn():
    await asyncio.sleep(5)


# TODO: Consider testable thread (for possible exceptions raised). See ref:
#       https://gist.github.com/sbrugman/59b3535ebcd5aa0e2598293cfa58b6ab
@pytest.fixture
def start_thread():

    thread_loop, thread = None, None

    def _start_thread(func):
        nonlocal thread_loop
        nonlocal thread
        thread_loop = asyncio.new_event_loop()
        thread = threading.Thread(target=func, args=(thread_loop,))
        thread.start()
        return thread_loop, thread

    yield _start_thread

    if thread_loop and thread:
        thread_loop.call_soon_threadsafe(thread_loop.stop)
        thread.join()
        thread_loop.close()


@pytest.mark.asyncio
async def test_cancellable_suppresses_error():

    async def run_and_cancel(coro_fn):
        task = asyncio.create_task(coro_fn())
        await asyncio.sleep(0)
        task.cancel()
        await task

    # Confirm CancelledError is raised normally
    with pytest.raises(asyncio.CancelledError):
        await run_and_cancel(async_fn)

    # Cancellable async fn doesn't raise when cancellable.
    await run_and_cancel(cancellable(async_fn))


def test_decorated_fn_still_async():
    assert inspect.iscoroutinefunction(cancellable_async_fn)


def test_decorating_non_async_fn_raises():

    with pytest.raises(ValueError, match="only decorate async functions"):

        @cancellable
        def sync_fn():
            pass


def test_cancel_loop_tasks_different_thread(start_thread):

    def thread_func(loop):
        asyncio.set_event_loop(loop)
        loop.run_until_complete(cancellable_async_fn())
        loop.run_forever()

    # Start a new thread with its own event loop running thread_func
    thread_loop, _ = start_thread(thread_func)
    cancel_loop_tasks(thread_loop)


# TODO: test cancel tasks
# TODO: test repeat



"""Mock server module class."""

import asyncio
import threading

from .async_utils import cancel_loop_tasks
from .connector import Connector


__all__ = ["MockConnector"]


class MockConnector(Connector):
    """Mock server, started on a separate thread."""

    def __init__(self, gui_id, verbose=False):
        self.gui_id = gui_id
        self.verbose = verbose
        self._clear()

    def _clear(self):
        self.__connector_thread = None
        self.__connector_loop = None

        # Connection related attributes
        self.__client_connected_event = threading.Event()

    def is_connected(self):
        """Indicates whether the mock 'connection' is established."""
        return self.__connector_thread.is_alive() is not None

    async def _mock_send_data(self, component):
        if self.verbose:
            print("Sending data:")
            print("Gui id:", self.gui_id)
            print("Component id:", component.id)
            print("data:", component.data)

    def send_component_data(self, component, require_connection=False):
        """Mocks sending data.

        If verbose=True was set in class, prints out what it would send.

        Args:
            - component: ComponentDataSender whose data will be sent over.
            - require_connection: (default: False) whether to raise an exception
                if there is no connection or to do nothing silently.

        Raises:
            RuntimeError: if require_connection is True and there are no
                connections at the time of being called.
        """
        if require_connection and not self.is_connected():
            raise RuntimeError("Not Connected!")
        future = asyncio.run_coroutine_threadsafe(
            self._mock_send_data(component), self.__connector_loop
        )
        future.result()

    def _mock_server_thread_fn(self, loop):

        async def _mock_start_server():
            self.__client_connected_event.set()
            if self.verbose:
                print("starting mock connector")

        asyncio.set_event_loop(loop)
        loop.run_until_complete(_mock_start_server())
        loop.run_forever()

    def wait_for_connection(self, timeout=None):
        """Wait for the mock 'connection' to be ready.

        Args:
            - timeout: (default None) The amount of time (in seconds) to wait
                before raising an error.

        Raises:
            TimeoutError: If timeout arg is set and the wait lasts for longer
                than the amount specified.
        """
        if not self.__client_connected_event.wait(timeout=timeout):
            raise TimeoutError("Timeout waiting for client connection")

    def start_connector_thread(self):
        """Start a mock 'server' on a new thread. It does nothing."""
        # Create a new event loop and start it in a new thread.
        self.__connector_loop = asyncio.new_event_loop()
        self.__connector_thread = threading.Thread(
            target=self._mock_server_thread_fn,
            args=(self.__connector_loop,),
        )
        self.__connector_thread.start()

    def stop_connector_thread(self):
        """Stops the mock 'server' and the thread it was on."""
        cancel_loop_tasks(self.__connector_loop)

        # Stop the event loop
        if self.__connector_loop and self.__connector_loop.is_running():
            self.__connector_loop.call_soon_threadsafe(self.__connector_loop.stop)
        self.__connector_thread.join()  # wait for thread to cleanup.
        self.__connector_loop.close()

        self._clear()



"""Core components submodule"""

from .base_component import BaseComponent, _BaseComponentMeta
from .data_sender_base_component import DataSenderBaseComponent
from .value_receiver_base_component import ValueReceiverBaseComponent
from .single_value_component import SingleValueComponent

__all__ = [
    "BaseComponent",
    "DataSenderBaseComponent",
    "ValueReceiverBaseComponent",
]


# Remove files which are added to the namespace because of imports but are not
# meant to be accessed directly by the user.
# pylint:disable=undefined-variable
# mypy: ignore-errors
del (
    base_component,
    data_sender_base_component,
    value_receiver_base_component,
    single_value_component,
)



"""SingleBooleanComponent class module."""

from .core import SingleValueComponent


__all__ = ["Toggle"]


class SingleBooleanComponent(SingleValueComponent):
    """Boolean receiver class which contains a single value."""

    def __init__(self, component_id):
        super().__init__(component_id, False)


class Toggle(SingleBooleanComponent):
    """Toggle class containing a single boolean value.

    The class has a single boolean value that can be accessed as an attribute:
    `toggle.value`.
    """



"""Mock server module class."""

import logging

from .connector import Connector


__all__ = ["MockConnector"]


class MockConnector(Connector):
    """Mock connector, started on a separate thread."""

    def __init__(self, gui_id):
        super().__init__()
        self.gui_id = gui_id
        self._logger = logging.getLogger(__name__)

    async def _send_component_data(self, component):
        self._logger.debug(
            "Sending data to gui '%s' and component '%s'",
            self.gui_id,
            component.id,
        )
        # TODO: Consider nicer way of logging this?
        self._logger.debug("data: %s", component.data)

    async def _start_connector(self):
        self.is_connected = True
        self._logger.debug("starting mock connector")

    async def _stop_connector(self):
        """Stops the mock 'connector'."""
        self._logger.debug("stopping mock connector")



import pytest

from .base_component import BaseComponent


class Component(BaseComponent):
    pass


def test_id_available():
    cid = "id1"
    c1 = Component(cid)

    assert c1.id is cid


def test_invalid_id_raises():
    with pytest.raises(
        ValueError, match=f"Component id must be a non empty string.*{None}"
    ):
        Component(None)
    with pytest.raises(
        ValueError, match=f"Component id must be a non empty string.*{9}"
    ):
        Component(9)
    with pytest.raises(ValueError, match="Component id must be a non empty string.*"):
        Component("")



"""ConnectorContext (alias connect) module.

This is the module that provides the easy-to-use ability for users to establish
a connection to the Mission Control app.
"""

import asyncio
import contextvars
import functools
import inspect
import logging


from .connector_config import config
from .mock_connector import MockConnector
from .websocket_server import WebSocketServer

__all__ = ["ConnectorContext", "context_connector"]


# Timeout (s) when waiting for a connection. Scripts running this code are meant
# to be run directly by the app that immediately starts the connection, so 5
# seconds should be more than enough.
CONNECTION_TIMEOUT = 5

# User friendly error message regarding timeouts when waiting for a connection.
# Users seeing this are likely trying to run the script that calls this code
# directly. This does not work and it should be called from the app. The error
# message should explain this and provide alternatives for debugging.
CONNECTION_TIMEOUT_ERROR_MSG = (
    "Unable to establish a connection with the Mission Control app. This is "
    "possibly because you are running your script directly instead of through "
    "the Mission Control app. If you need to do so for debugging purposes, use "
    "`zmc.connect(debug=True)` to mock out the connection."
)

# User friendly error message regarding multiple connections. Users seeing this
# must be nesting their context or calls to decorated functions.
MULTIPLE_CONNECTIONS_ERROR_MSG = (
    "Cannot create multiple connections at the same time. This is likely "
    "happening because of nested `with` statements or nested calls to "
    "functions decorated with `zmc.connect()`."
)

# Context var containing the Connector object that the context manager starts.
context_connector = contextvars.ContextVar("context_connector", default=None)

_root_logger = logging.getLogger()


class ConnectorContext:
    """Context manager for a connection to the Mission Control app.

    Starts a connection in a separate thread to the one that the manager is
    called in. The context won't start until a connection has been established.
    If no connection is established before the timeout, a RuntimeError will be
    raised.

    Usage as a context manager:
    ```
    import zmc

    with zmc.connect():
        do_something()  # will execute while connection is established.
    ```

    This class supports both synchronous and asynchronous context management. In
    asynchronous contexts, the wait for the connection is done asynchronously so
    it does not prevent other tasks from executing during the wait. It will,
    however, still prevent the context from entering until a connection is made.

    For convenience, instances of this class can also be used as decorators for
    functions. Any function decorated will, when called, first enter a context
    and then execute.

    Usage as decorator:
    ```
    import zmc

    @zmc.connect()
    def do_something():
        ... # arbitrary implementation

    do_something()  # will execute while connection is established.
    ```

    To run the script without requiring a connection to the app, use the
    `debug=True` argument. This will mock out the connection with the Mission
    Control app. No values will be read or sent from the app but the rest of the
    code will work as expected.

    Example:
    ```
    with zmc.connect(debug=True):
        do_something()
    ```

    args: (keyword-only)
        - debug: (default: False). Whether or not to enter the context in
                debugging mode. If true, the connection will be mocked.
    """

    def __init__(self, *, debug=False, verbose=False):
        self._verbose = verbose
        self._prev_logger_level = None
        self._connector = (
            WebSocketServer("localhost", config.port, config.gui_id)
            if not debug
            else MockConnector(config.gui_id)
        )
        self.contextvar_token = None

    def __call__(self, func):
        # Async case. It is almost identical to the sync one but with a few
        # added "async" and "await" keywords.
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def aconnector_wrapper(*args, **kwargs):
                async with self:
                    await func(*args, **kwargs)

            return aconnector_wrapper

        # Sync case
        @functools.wraps(func)
        def connector_wrapper(*args, **kwargs):
            with self:
                func(*args, **kwargs)

        return connector_wrapper

    def _start_connector_thread(self):
        if context_connector.get() is not None:
            raise RuntimeError(MULTIPLE_CONNECTIONS_ERROR_MSG)
        self.contextvar_token = context_connector.set(self._connector)
        self._connector.start_connector_thread()

    def __enter__(self):
        if self._verbose:
            self._prev_logger_level = _root_logger.level
            _root_logger.setLevel(logging.DEBUG)
        try:
            self._start_connector_thread()
            # Await a connection. Otherwise the zmc code inside the context
            # won't work as intended.
            self._connector.wait_for_connection(CONNECTION_TIMEOUT)
        # TODO: cleaner way of handling exceptions and always exiting?
        except TimeoutError as exc:
            self.__exit__(None, None, None)
            raise RuntimeError(CONNECTION_TIMEOUT_ERROR_MSG) from exc
        except Exception:
            self.__exit__(None, None, None)
            raise
        except SystemExit:
            self.__exit__(None, None, None)

    async def __aenter__(self):
        if self._verbose:
            self._prev_logger_level = _root_logger.level
            _root_logger.setLevel(logging.DEBUG)
        try:
            self._start_connector_thread()
            # Await a connection. Otherwise the zmc code inside the context
            # won't work as intended. Given that this is an async context, the
            # wait happens on another thread to not block other async actions.
            # It will still, however, block the context from entering until the
            # connection happens.
            await asyncio.get_event_loop().run_in_executor(
                None, self._connector.wait_for_connection, CONNECTION_TIMEOUT
            )
        except TimeoutError as exc:
            await self.__aexit__(None, None, None)
            raise RuntimeError(CONNECTION_TIMEOUT_ERROR_MSG) from exc
        except Exception:
            await self.__aexit__(None, None, None)
            raise
        except SystemExit:
            await self.__aexit__(None, None, None)

    # pylint: disable=inconsistent-return-statements
    def __exit__(self, exc_type, exc_value, traceback):
        self._connector.stop_connector_thread()
        context_connector.reset(self.contextvar_token)

        # reset logging level that was being used before entering
        if self._prev_logger_level:
            _root_logger.setLevel(self._prev_logger_level)

        # If an exception was raised in the connector thread, raise that now.
        if self._connector.has_thread_exception():
            self._connector.raise_thread_exception()

        if exc_type is SystemExit:
            return True  # Suppress the SystemExit exception

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.__exit__(exc_type, exc_value, traceback)



"""Utils package for zmc library."""

from .singleton import singleton
from .instance_registry import InstanceRegistry
from .logger import *

# Remove files which are added to the namespace because of imports but are not
# meant to be accessed directly by the user.
# pylint:disable=undefined-variable
# mypy: ignore-errors
del instance_registry, logger



"""Connector config class module."""

import os

from zmc.utils import singleton

__all__ = ["config"]


@singleton
class ConnectorConfig:
    """Config class used within the connector context.

    The config reads values from the os environment and provides defaults if the
    env variables are not available. The config looks for two values:
        - ZMC_GUI_ID: (default 'gui-id') id of the gui that the script should
            connect to.
        - ZMC_PORT: (default 8766) the port a server should be serving and
            listening to.
    """

    def __init__(self):
        self.gui_id = os.getenv("ZMC_GUI_ID", "gui-id")
        self.port = int(os.getenv("ZMC_PORT", "8766"))


config = ConnectorConfig()



"""SingleNumberComponent class module."""

from .single_value_component import SingleValueComponent


__all__ = ["SingleNumberComponent"]


class SingleNumberComponent(SingleValueComponent):
    """Number receiver class which contains a single value."""

    def __init__(self, component_id):
        super().__init__(component_id, 1)



"""ButtonComponent class module."""

from .core import ValueReceiverBaseComponent


__all__ = ["Button"]


class Button(ValueReceiverBaseComponent):
    """Button class that calls a function whenever it is clicked."""

    # No value but this is still needed to make the class concrete.
    def _set_value(self, _):
        """No value is set for buttons"""



"""Abstract connector class module."""

import asyncio
import ctypes
import threading

from abc import ABC, abstractmethod

from .async_utils import cancel_tasks_in_loop, cancellable

# User friendly error message about there not being a connection when the user
# tries to send data over to the Mission Control app.
SEND_DATA_NOT_CONNECTED_ERROR_MSG = (
    "Tried to send data to the Mission Control app but the connection has been "
    "lost. If you don't see an error in the app, then this is likely due to an "
    "internal issue. Try restarting it to see whether it resolves itself."
)


class Connector(ABC):
    """Abstract class fpr starting and stopping a connector on a separate thread

    This base class takes care of all of the threading logic. To implement it,
    subclasses need to implement 3 async methods:
        - _send_component_data(component)
            -  Args: component: DataSenderComponent whose data will be sent over
        - _start_connector()
        - _stop_connector()
    """

    def __init__(self):
        self.__original_thread = None
        self.__connector_thread = None
        self.__connector_thread_exception = None
        self.__connector_loop = None
        self.__connected_event = threading.Event()

    def __reset(self):
        self.__original_thread = None
        self.__connector_thread = None
        # Note: self.__connector_thread_exception is purposefully not cleared.
        #       It needs to remain so that callers can bubble it up.
        self.__connector_loop = None
        self.__connected_event.clear()

    @property
    def is_connected(self):
        """Indicates whether the connection has been established."""
        return self.__connected_event.is_set()

    @is_connected.setter
    def is_connected(self, connected):
        """Whether the Connector is connected to a client."""
        if connected:
            self.__connected_event.set()
        else:
            # reset connected event
            self.__connected_event = threading.Event()

    @property
    def _original_thread(self):
        return self.__original_thread

    def wait_for_connection(self, timeout=None):
        """Wait for the connection to be established.

        Once a connection has been made, this will return any calls immediately.
        This will keep being the case until the connector is marked as
        disconnected.

        Args:
            - timeout: (default None) The amount of time (in seconds) to wait
                before raising an error.

        Raises:
            TimeoutError: If timeout arg is set and the wait lasts for longer
                than the amount specified.
        """
        if not self.__connected_event.wait(timeout=timeout):
            raise TimeoutError("Timeout waiting for connection")

        if self.__connector_thread_exception:
            # If there was an exception on the thread, then actually there is no
            # connection and it was set to True simply to end the wait. In that
            # case, we reset it now just for correctness.
            self.is_connected = False

    async def _run_coroutine_fn(self, coro_fn, *args, **kwargs):
        try:
            await cancellable(coro_fn)(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._handle_thread_exception(e)

    @abstractmethod
    async def _send_component_data(self, component):
        """Send component data over to the mission control app."""

    def send_component_data(self, component, require_connection=False):
        """Send data over to a given component.

        Args:
            - component: DataSenderComponent whose data will be sent over.
            - require_connection: (default: False) whether to raise an exception
                if there is no connection or to do nothing silently.

        Raises:
            RuntimeError: if require_connection is True and there are no
                connections at the time of being called.
        """
        if require_connection and not self.is_connected:
            raise RuntimeError(SEND_DATA_NOT_CONNECTED_ERROR_MSG)
        asyncio.run_coroutine_threadsafe(
            self._run_coroutine_fn(self._send_component_data, component),
            self.__connector_loop,
        )

    @abstractmethod
    async def _start_connector(self):
        """Async function that starts the connector.

        This is an async function that should return immediately after starting
        the connector. It should not wait as long as the connection is open
        (e.g. a server should not call `await server.wait_closed()`)
        """

    def _connector_thread_fn(self, loop, connector_ready_event):
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(
                self._run_coroutine_fn(self._start_connector)
            )
            connector_ready_event.set()
        finally:
            loop.run_forever()

    def start_connector_thread(self):
        """Start the connector on a new separate thread.

        The thread that the connector was started from is stored and accessible
        via `self._original_thread`. The function does not return until the
        connector has been fully started.
        """
        if self.__connector_thread_exception:
            # If there is a thread exception at the start, then it is from a
            # previous attempt. We clear it here so that we are starting fresh.
            self.__connector_thread_exception = None

        self.__original_thread = threading.current_thread()

        def handle_thread_exception(_loop, context):
            exception = context.get("exception", context["message"])
            self._handle_thread_exception(exception)

        # Create a new event loop.
        self.__connector_loop = asyncio.new_event_loop()
        self.__connector_loop.set_exception_handler(handle_thread_exception)
        # Start the event loop in a new thread.
        connector_ready_event = threading.Event()
        self.__connector_thread = threading.Thread(
            target=self._connector_thread_fn,
            args=(self.__connector_loop, connector_ready_event),
        )
        self.__connector_thread.start()

        # Wait until the connector is ready before returning
        connector_ready_event.wait()

    @abstractmethod
    async def _stop_connector(self):
        """Async function used to stop the connector."""

    def _stop_thread(self, num_retries=None, timeout=0.1):
        if not self.__connector_thread.is_alive():
            self.__connector_loop.close()
            return  # Connector thread has stopped and loop is closed. Hooray!

        if num_retries < 0:
            raise RuntimeError("Unable to properly close connection thread.")

        # Try canceling all tasks in the loop and and resolving the thread
        cancel_tasks_in_loop(self.__connector_loop)
        if self.__connector_loop and self.__connector_loop.is_running():
            self.__connector_loop.call_soon_threadsafe(
                self.__connector_loop.stop
            )
        # Wait for thread to cleanup. If it doesn't within the timeout, move on.
        self.__connector_thread.join(timeout=timeout)

        # Verify that the thread is stopped or retry if it is not.
        self._stop_thread(
            num_retries=num_retries - 1 if num_retries is not None else None,
            timeout=timeout,
        )

    def stop_connector_thread(self):
        """Stop the connector and the thread it is on.

        After stopping the connector, all other tasks on the connector thread
        loop are also canceled. Then the thread itself is closed and the class
        instance is reset.
        """
        # Stop the connector on the connector thread.
        future = asyncio.run_coroutine_threadsafe(
            self._stop_connector(), self.__connector_loop
        )
        future.result()

        self._stop_thread(num_retries=5)

        self.__reset()

    def _exit_original_thread(self):
        if self._original_thread is None:
            return  # Nothing to do if we don't know which thread to exit.
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(self._original_thread.ident),
            ctypes.py_object(SystemExit),
        )

    def _handle_thread_exception(self, exception):
        # Store the exception which users can access later.
        self.__connector_thread_exception = exception
        # Stop the original thread by sending an interrupt signal.
        self._exit_original_thread()
        # This is a bit of a misapplication but if the error is raised during a
        # wait for connection, we want the program to stop waiting since there
        # is nothing to wait for. So we set is_connected to True (this will be
        # reset immediately after the wait ends). Then the client can handle the
        # error however they choose.
        self.is_connected = True

    def has_thread_exception(self):
        """Whether an exception was raised in the connector thread."""
        return self.__connector_thread_exception is not None

    def raise_thread_exception(self):
        """Raises the exception from the connector thread, if any."""
        if self.has_thread_exception():
            raise self.__connector_thread_exception from None



"""WebsocketServer class module."""

import asyncio
import ctypes
import json
import sys
import threading
import traceback
import websockets

from ..components.value_receiver_base_component import ValueReceiverBaseComponent
from ..components.component_instance_registry import ComponentInstanceRegistry
from ..version import __version__

from .async_utils import cancellable, cancel_loop_tasks, async_repeat
from .connector import Connector


__all__ = ["WebSocketServer"]


# TODO: This is not a very helpful message for users. It's good enough for now
#       but at some point we'll want to fix that (which component, id, and
#       details as to what is going on are unclear).
# User friendly error message regarding the fact that a component that does not
# allow for setting a value was given an idea that corresponds to a control in
# the Mission Control app.
SETTING_VALUE_ON_NON_SETTER_ERROR_MSG = (
    "Component id was attached to the wrong component. Specifically, it was "
    "set on a component that does not accept a value."
)

# Rate (in Hz) at which messages will be sent over websockets.
_THROTTLE_RATE = 61


# TODO: handle errors in handler?
# TODO: test interrupt logic (exit) on windows, then cleanup
class WebSocketServer(Connector):
    """Server that allows caller to open up a websocket on a separate thread.

    Args:
        - host: host as per similarly named asyncio.serve() function parameter
        - port: port as per similarly named asyncio.serve() function parameter
        - gui_id: id of the gui this server is going to connect to.
    """

    def __init__(self, host, port, gui_id):
        self.host = host
        self.port = port
        self.gui_id = gui_id
        self._clear()
        self._throttle_rate = 60

    def _clear(self):
        # Original caller related attributes
        self.__original_thread = None

        # Server related attributes
        self.__server = None
        self.__server_thread = None
        self.__server_loop = None

        # Connection related attributes
        self.__client_connected_event = threading.Event()
        self.__websockets = set()

        # Queued data related attirbutes
        self.__data_component_queue = asyncio.Queue()
        self.__send_queued_data_task = None

    def is_serving(self):
        """Indicates whether the server is currently up and running."""
        return self.__server is not None and self.__server.is_serving()

    def is_connected(self):
        """Indicates whether there is at least one websocket connection."""
        return self.is_serving() and self.__websockets

    async def _exit_original_thread(self):
        if self.__original_thread is None:
            return  # Nothing to do if we don't know which thread to exit.
        # TODO: test on windows, then cleanup
        # signal.pthread_kill(self.original_thread.ident, signal.SIGUSR1)
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(self.__original_thread.ident), ctypes.py_object(SystemExit)
        )

    @cancellable
    async def _send_all_component_data(self, components):
        data_event = {
            "type": "send-all-data",
            "guiId": self.gui_id,
            "components": [
                {
                    "componentId": component.id,
                    "data": component.data,
                }
                for component in components.values()
            ],
            "pyVersion": __version__,
        }
        json_data_event = json.dumps(data_event)
        websockets_set = self.__websockets.copy()
        for websocket in websockets_set:
            try:
                await websocket.send(json_data_event)
            except websockets.ConnectionClosed:
                pass

    @cancellable
    @async_repeat(1 / _THROTTLE_RATE)
    async def _send_queued_component_data(self):
        items = {}
        while not self.__data_component_queue.empty():
            component = await self.__data_component_queue.get()
            items[component.id] = component
        if items:
            await self._send_all_component_data(items)

    async def _start_send_queued_data_task(self):
        if self.__send_queued_data_task is None or self.__send_queued_data_task.done():
            self.__send_queued_data_task = asyncio.create_task(
                self._send_queued_component_data()
            )

    async def _queue_component_data(self, component):
        await self.__data_component_queue.put(component)
        await self._start_send_queued_data_task()

    def send_component_data(self, component, require_connection=False):
        """Send data over to a given component.

        Args:
            - component: DataSenderComponent whose data will be sent over.
            - require_connection: (default: False) whether to raise an exception
                if there is no connection or to do nothing silently.

        Raises:
            RuntimeError: if require_connection is True and there are no
                connections at the time of being called.
        """
        if require_connection and not self.is_connected():
            raise RuntimeError("Not Connected!")
        asyncio.run_coroutine_threadsafe(
            self._queue_component_data(component), self.__server_loop
        )

    # TODO: should refactor this to be a little more palettable
    async def _websocket_consumer(self, websocket):
        async for message in websocket:
            event = json.loads(message)
            # TODO: handle app versions (for future proofing as things change)
            _app_version = event["app_version"]
            if event["type"] == "set-value":
                target_id = event["component_id"]
                component = ComponentInstanceRegistry.get_instance(target_id)
                if component is None:
                    continue  # Didn't find component. Do nothing and skip.
                if not isinstance(component, ValueReceiverBaseComponent):
                    raise ValueError(SETTING_VALUE_ON_NON_SETTER_ERROR_MSG)
                component.receive_value(event.get("value", None))
            if event["type"] == "initialize":
                components = event["components"]
                for component_json in components:
                    component_id = component_json["component_id"]
                    component = ComponentInstanceRegistry.get_instance(component_id)
                    if component is None:
                        continue  # Didn't find component. Do nothing and skip.
                    value = component_json.get("value", None)
                    if not value:
                        continue  # No value to update.
                    if not isinstance(component, ValueReceiverBaseComponent):
                        raise ValueError(SETTING_VALUE_ON_NON_SETTER_ERROR_MSG)
                    component.receive_value(value)
                self.__client_connected_event.set()  # Signal that a client is ready
            if event["type"] == "kill-func":
                self._exit_original_thread()

    async def _websocket_handler(self, websocket, _):
        self.__websockets.add(websocket)
        consumer_task = asyncio.create_task(self._websocket_consumer(websocket))

        done, pending = await asyncio.wait(
            [consumer_task], return_when=asyncio.FIRST_COMPLETED
        )

        for task in pending:
            task.cancel()

        self.__websockets.discard(websocket)

        # TODO: handle errors? Not quite right. This throws an error and prints
        #       it but doesn't do anything else. We would need to propagate
        #       somehow or be happy with this behavior. Worth looking into more.
        for task in done:
            try:
                error = task.exception()
                if error is not None:
                    # For now, simply print out the error
                    traceback.print_exception(
                        type(error), error, error.__traceback__, file=sys.stdout
                    )
            except asyncio.CancelledError:
                pass

    @cancellable
    async def _run_server(self, server_ready_event):
        self.__server = await websockets.serve(
            self._websocket_handler, self.host, self.port
        )
        server_ready_event.set()  # Signal that the server is ready
        await self.__server.wait_closed()

    def _stop_server(self):
        if not self.is_serving():
            return  # Server is already not running so nothing to do.

        async def stop_server():
            self.__server.close()
            await self.__server.wait_closed()

        future = asyncio.run_coroutine_threadsafe(stop_server(), self.__server_loop)
        future.result()

    def _server_thread_fn(self, loop, server_ready_event):
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(self._run_server(server_ready_event))
        finally:
            loop.run_forever()

    def wait_for_connection(self, timeout=None):
        """Wait for the first client websocket connection to be made.

        Once a first connection has been made, this will return immediately
        for the rest of the time the server is up. This is true even if all of
        the connections have since been closed.

        Args:
            - timeout: (default None) The amount of time (in seconds) to wait
                before raising an error.

        Raises:
            TimeoutError: If timeout arg is set and the wait lasts for longer
                than the amount specified.
        """
        if not self.__client_connected_event.wait(timeout=timeout):
            raise TimeoutError("Timeout waiting for client connection")

    def start_connector_thread(self):
        """Start the websocket server on a new separate thread.

        The thread that the server was started from is stored so that it can be
        exited upon a request via the websocket. The function does not return
        until the server has fully been started.
        """
        self.__original_thread = threading.current_thread()

        # TODO: change this to interrupt the main thread?
        def handle_thread_exception(_loop, context):
            msg = context.get("exception", context["message"])
            print(f"Caught exception: {msg}")

        # Create a new event loop and start it in a new thread.
        self.__server_loop = asyncio.new_event_loop()
        self.__server_loop.set_exception_handler(handle_thread_exception)
        server_ready_event = threading.Event()
        self.__server_thread = threading.Thread(
            target=self._server_thread_fn,
            args=(self.__server_loop, server_ready_event),
        )
        self.__server_thread.start()

        # Wait until the server is ready before returning
        server_ready_event.wait()

    def _stop_thread(self, num_retries=None, timeout=0.1):
        if not self.__server_thread.is_alive():
            self.__server_loop.close()
            return  # Server thread has stopped and the loop is closed. Hooray!
        if num_retries < 0:
            raise RuntimeError("Unable to properly close connection thread.")
        cancel_loop_tasks(self.__server_loop)
        if self.__server_loop and self.__server_loop.is_running():
            self.__server_loop.call_soon_threadsafe(self.__server_loop.stop)
        self.__server_thread.join(timeout=timeout)  # wait for thread to cleanup.
        self._stop_thread(
            num_retries=num_retries - 1 if num_retries is not None else None,
            timeout=timeout,
        )

    def stop_connector_thread(self):
        """Stop the websocket server and the thread it was on.

        After stopping the server, all other tasks on the server thread loop are
        also canceled and then the thread itself is closed and the class
        instance is reset.
        """
        self._stop_server()
        self._stop_thread(num_retries=5)

        self._clear()



"""SingleBooleanComponent class module."""

from .single_value_component import SingleValueComponent


__all__ = ["SingleBooleanComponent"]


class SingleBooleanComponent(SingleValueComponent):
    """Boolean receiver class which contains a single value."""

    def __init__(self, component_id):
        super().__init__(component_id, False)



"""SingleBooleanComponent class module."""

from .core import BaseComponent
from .core import SingleValueComponent

from .single_number_components import (
    NumericInput,
    SingleNumberComponent,
    Slider,
)


def test_is_single_value_subclass():
    assert issubclass(SingleNumberComponent, SingleValueComponent)


def test_default_value():
    c = SingleNumberComponent("id")

    assert c.value == 1


def test_subclasses():
    assert issubclass(Slider, SingleNumberComponent)
    assert issubclass(NumericInput, SingleNumberComponent)


def test_gets_registered():
    id1 = "id1"
    id2 = "id2"
    c1 = Slider(id1)
    assert BaseComponent.registry.get(id1) == c1
    c2 = NumericInput(id2)
    assert BaseComponent.registry.get(id2) == c2



"""Abstract connector class module."""

from abc import ABC, abstractmethod


# TODO: move a large portion of the logic here. I think the majority of the
#       separate thread stuff can be moved here. Same goes for connection ready
#       logic.
class Connector(ABC):
    """Class that allows starting and stopping a server on a separate thread."""

    @abstractmethod
    def is_connected(self):
        """Indicates whether the connection has been established."""

    @abstractmethod
    def send_component_data(self, component, require_connection=False):
        """Send component data over to the mission control app."""

    @abstractmethod
    def start_connector_thread(self):
        """Start the server on a new separate thread."""

    @abstractmethod
    def stop_connector_thread(self):
        """Stop the server and the thread that it is running."""

    @abstractmethod
    def wait_for_connection(self, timeout=None):
        """Wait for the first client connection (with an optional timeout)."""



"""Component implementations submodule.

Any component implementation should be exposed through here.
"""

from ..component_instance_registry import ComponentInstanceRegistry

from .graph_component import GraphComponent as Graph
from .single_boolean_component import SingleBooleanComponent
from .single_text_component import SingleTextComponent
from .single_number_component import SingleNumberComponent
from .button_component import ButtonComponent as Button


__all__ = ["Button", "Graph", "NumericInput", "Slider", "TextInput", "Toggle"]


@ComponentInstanceRegistry
class Slider(SingleNumberComponent):
    """Slider class which contains a single value.

    A slider has a value of a single number that can be accessed as an
    attribute: `slider.value`.
    """


@ComponentInstanceRegistry
class NumericInput(SingleNumberComponent):
    """Numeric class, representing a freeform or dropdown number component.

    The class has a single value that can be accessed as an attribute:
    `numeric_input.value`.
    """


@ComponentInstanceRegistry
class TextInput(SingleTextComponent):
    """Text class, representing a freeform or dropdown text component.

    The class has a single value that can be accessed as an attribute:
    `text_input.value`.
    """


@ComponentInstanceRegistry
class Toggle(SingleBooleanComponent):
    """Toggle class containing a single boolean value.

    The class has a single boolean value that can be accessed as an attribute:
    `toggle.value`.
    """



"""BaseComponentValueSetter class module."""

import inspect

from abc import abstractmethod

from .base_component import BaseComponent


__all__ = ["ValueReceiverBaseComponent"]


# TODO: implement `remove_callback`
# TODO: factor out into Callback function? might even already exist in py
# TODO: pass in a value as one of the arguments?
class ValueReceiverBaseComponent(BaseComponent):
    """Base class for any component that can set a value.

    Any concrete subclass must implement _set_value(self, value).
    """

    def __init__(self, component_id):
        super().__init__(component_id)
        self._func = None
        self._func_args = []
        self._func_kwargs = {}

    def set_callback(self, func, *args, **kwargs):
        """Sets a callback function that will be called when value is received.

        Any arguments passed in after the callback function will be passed to
        the function when it is called.
        """
        if not inspect.isroutine(func):
            raise ValueError(
                f"`func` must be a callable function, given: {func}"
            )

        self._func = func
        self._func_args = args
        self._func_kwargs = kwargs

    @abstractmethod
    def _set_value(self, value):
        """Store the value pass in for future access."""

    def receive_value(self, value):
        """Receive the value sent from the Mission Control app.

        First sets the value on the component. Once that is done, calls any
        callback functions that have been added.

        Args:
            value: component value that has been received.
        """
        self._set_value(value)
        if self._func:
            self._func(*self._func_args, **self._func_kwargs)



"""SingleTextComponent class module."""

from .single_value_component import SingleValueComponent


__all__ = ["SingleTextComponent"]


class SingleTextComponent(SingleValueComponent):
    """Text receiver class which contains a single value."""

    def __init__(self, component_id):
        super().__init__(component_id, "")



"""Basic logger configuration submodule"""

import logging
import sys

__all__ = []

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s]: %(message)s",
    handlers=[logging.StreamHandler(sys.stdout)],
)



"""BaseComponent class module."""

from abc import ABCMeta

from zmc.utils import InstanceRegistry

__all__ = ["BaseComponent"]


class _BaseComponentMeta(ABCMeta):
    """Metaclass for BaseComponent class which contains an instance registry."""

    __REGISTRY = InstanceRegistry()

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)

    @property
    def registry(cls):
        """Get the instance registry dict for the BaseComponent class"""
        return _BaseComponentMeta.__REGISTRY.get_dict()

    def __call__(cls, *args, **kwds):
        instance = super().__call__(*args, **kwds)
        _BaseComponentMeta.__REGISTRY.register(instance)
        return instance


class BaseComponent(metaclass=_BaseComponentMeta):
    """Base class for all components.

    The base component class simply contains a unique id that can be accessed as
    a property.
    """

    def __init__(self, component_id):
        if not isinstance(component_id, str) or component_id == "":
            raise ValueError(
                "Component id must be a non empty string, given:", component_id
            )
        self.__id = component_id

    @property
    def id(self) -> str:
        """Unique identifier for component."""
        return self.__id



from .core import BaseComponent
from .core import ValueReceiverBaseComponent

from .button_component import Button


def test_is_receiver_subclass():
    assert issubclass(Button, ValueReceiverBaseComponent)


def test_is_concrete():
    Button("id")


def test_gets_registered():
    cid = "id"
    c = Button(cid)
    assert BaseComponent.registry.get(cid) == c


# TODO: test callbacks



"""BaseComponent class module."""

from abc import ABC

__all__ = ["BaseComponent"]


# pylint: disable=too-few-public-methods
class BaseComponent(ABC):
    """Base class for all components.

    The base component class simply contains a unique id that can be accessed as
    a property.
    """

    def __init__(self, component_id):
        if not isinstance(component_id, str) or component_id == "":
            raise ValueError(
                "Component id must be a non empty string, given:", component_id
            )
        self.__id = component_id

    @property
    def id(self) -> str:
        """Unique identifier for component."""
        return self.__id



"""SingleValueComponent class module."""

from .value_receiver_base_component import ValueReceiverBaseComponent


__all__ = ["SingleValueComponent"]


class SingleValueComponent(ValueReceiverBaseComponent):
    """Receiver class which contains a single value."""

    def __init__(self, component_id, default_value):
        super().__init__(component_id)
        self.__value = default_value

    @property
    def value(self):
        """Value stored by the class."""
        return self.__value

    def _set_value(self, value):
        self.__value = value["value"]
